# Copyright 1999-2017. Parallels IP Holdings GmbH. All Rights Reserved.
import subprocess
from contextlib import contextmanager
import logging as log
from collections import defaultdict
import re
from distutils.version import StrictVersion
import yum
import yum.Errors

from plesk_file_utils import discard_stdout
from os_detect import OS
import plesk_constants
import pum_errors


def importGpgKeyCallback(keyinfo):
    return keyinfo["keyurl"] and keyinfo["keyurl"].startswith("file://")


class PumYumBase(yum.YumBase):
    """ Override _checkSignatures to allow importing only locally stored gpg keys. """
    def _checkSignatures(self, pkgs, callback):
        ''' Check the signatures of the downloaded packages '''
        # This can be overloaded by a subclass.
        for po in pkgs:
            result, errmsg = self.sigCheckPkg(po)
            if result == 0:
                # Verified ok, or verify not req'd
                continue
            elif result == 1:
                self.getKeyForPackage(po, fullaskcb=importGpgKeyCallback)
            else:
                raise yum.Errors.YumGPGCheckError, errmsg
        return 0


class YumWrapper:
    def __init__(self, yum_base, version_compare, unsafe):
        self.yum = yum_base
        self.version_compare = version_compare
        self.pm = "yum"
        self._unsafe = unsafe
        # exclude kernel packages due to PPP-20268
        self._default_exclude = ["kernel", "kernel-*"]
        if not self._unsafe:
            self.yum.conf.exclude = self._default_exclude

    @staticmethod
    def _print_ver(epoch, version, release):
        """returns a printable version string - including epoch, if it's set (like yum does)"""
        if epoch != '0':
            ver = '%s:%s-%s' % (epoch, version, release)
        else:
            ver = '%s-%s' % (version, release)

        return ver

    @classmethod
    def _get_package_info(cls, package):
        return {
            "name": package.name,
            "arch": package.arch,
            "full_version": cls._print_ver(package.epoch, package.version, package.release),
        }

    @classmethod
    def _installed_pkg_repo(cls, package):
        """ Get repository source id of installed package.
            None if source is unknown.
        """
        return package.yumdb_info.get("from_repo")

    def get_installed_packages(self):
        """
        Retrieve the list of the installed packages.

        @return dictionary of (package name, package info), where package info
                is extended with repository id and summary
        """
        installed_packages = {}

        for package in self.yum.doPackageLists("installed").installed:
            package_info = self._get_package_info(package)
            package_info["repo_from"] = self._installed_pkg_repo(package)
            package_info["summary"] = package.summary

            installed_packages[package.name] = package_info

        return installed_packages

    @staticmethod
    def _extract_plesk_repo_version(repo):
        match = re.search(r'\d+_\d+_\d+', repo)
        if not match:
            return None
        return StrictVersion(match.group(0).replace("_", "."))

    @staticmethod
    def _is_centos_repo(repo):
        centos_repos = [None, "base", "updates"]
        return repo in centos_repos

    @staticmethod
    def _is_redhat_repo(repo):
        return repo is None or repo.startswith("rhel-")

    @staticmethod
    def _is_cloudlinux_repo(repo):
        return repo is None or repo.startswith("cloudlinux-")

    @staticmethod
    def _is_vzlinux_repo(repo):
        return repo is None or repo.startswith("virtuozzolinux-")

    @classmethod
    def _check_repo_compatibility(cls, repo1, repo2, unsafe=False):
        if unsafe:
            return True

        if repo1 == repo2:
            return True

        if cls._is_centos_repo(repo1) and cls._is_centos_repo(repo2):
            return True

        if cls._is_redhat_repo(repo1) and cls._is_redhat_repo(repo2):
            return True

        if cls._is_cloudlinux_repo(repo1) and cls._is_cloudlinux_repo(repo2):
            return True

        if cls._is_vzlinux_repo(repo1) and cls._is_vzlinux_repo(repo2):
            return True

        # Special case for anaconda:
        if ((repo1 and repo1.startswith("anaconda") and
             (cls._is_centos_repo(repo2) or cls._is_redhat_repo(repo2) or cls._is_cloudlinux_repo(repo2) or cls._is_vzlinux_repo(repo2)))):
            return True

        if not repo1 and repo2 and repo2.startswith("PLESK_"):
            ver2 = YumWrapper._extract_plesk_repo_version(repo2)
            if not ver2:
                # addon repository built by Plesk since it starts with PLESK_ and doesnot contain 3number version
                return True
            curver = StrictVersion(plesk_constants.PRODVERSION)
            if ver2 <= curver:
                return True

        if repo1 and repo1.startswith("PLESK_") and repo2 and repo2.startswith("PLESK_"):
            ver1 = YumWrapper._extract_plesk_repo_version(repo1)
            ver2 = YumWrapper._extract_plesk_repo_version(repo2)
            if not (ver1 and ver2):
                return False
            curver = StrictVersion(plesk_constants.PRODVERSION)
            if ver1 <= ver2 and ver2 <= curver:
                return True

        return False

    @classmethod
    def _check_arch_compatibility(cls, arch1, arch2):
        def arch_bits(arch):
            if arch == "noarch":
                return 0
            return 64 if "64" in arch else 32
        if arch2 == "noarch":
            return True
        if arch1 == "noarch" and arch_bits(arch2) == OS.bits():
            return True
        if arch_bits(arch1) == arch_bits(arch2):
            return True
        return False

    def exclude_packages(self, packages):
        self.yum.conf.exclude = packages
        if not self._unsafe:
            self.yum.conf.exclude += self._default_exclude

    def _prepare_update(self, packages):
        avail_dict = defaultdict(list)
        pl = self.yum.doPackageLists(showdups=True)

        for p in pl.available:
            avail_dict[p.name].append(p)

        ts_packages = []

        for pi in pl.installed:
            if (packages and pi.name not in packages) or pi.name not in avail_dict:
                continue
            candidate = None
            for pa in avail_dict.get(pi.name, []):
                if ((self._check_repo_compatibility(self._installed_pkg_repo(pi), pa.repoid, self._unsafe) and
                     self._check_arch_compatibility(pi.arch, pa.arch) and
                     ((candidate is None and pa.verGT(pi)) or (candidate is not None and pa.verGT(candidate))))):
                    candidate = pa
            if candidate:
                # TODO: replace specifying version with specifying target po
                self.yum.update(po=candidate)
                if not self._resolve_deps("Broken transaction adding package", raise_ex=False):
                    # resolving transaction failed: let's fix it
                    remove_keys = self.yum.tsInfo.pkgdict.keys()
                    for k in remove_keys:
                        self.yum.tsInfo.remove(k)
                    # now transaction should be empty
                    self._resolve_deps("Unable to cleanup transaction")

                    # return "good" packages back into transaction:
                    for k in ts_packages:
                        self.yum.update(po=k)
                else:
                    ts_packages.append(candidate)

        self._resolve_deps("Unable to resolve transaction")

    def _resolve_deps(self, message, raise_ex=True):
        (code, result) = self.yum.resolveDeps()
        # 0: Success - empty transaction
        # 2: Success - deps resolved
        if code in (0, 2):
            return True
        if raise_ex:
            raise Exception("%s (%s): %s" % (message, code, result))
        log.warning("%s (%s): %s", message, code, result)
        return False

    def simulate_update(self, packages):
        """
        Perform a simulation of update of the given packages.

        @param packages list of packages (names) to simulate an update of
        @return dictionary of (package name, updated package info) -- value type will be refined
        """
        transaction_info = {}

        self._prepare_update(packages)

        for package in self.yum.tsInfo.getMembers():
            if package.ts_state == 'ud':
                continue
            item = self._get_package_info(package)
            if len(package.updates) > 0:
                u = package.updates[0]
                item["from"] = self._print_ver(u.epoch, u.version, u.release)
                item["summary"] = package.updates[0].summary
                item["repo_from"] = self._installed_pkg_repo(package.updates[0])
                item["repo_to"] = package.repoid
            else:
                item["from"] = None
                item["summary"] = package.po.summary
                item["repo_from"] = None
                item["repo_to"] = package.repoid
            transaction_info[package.name] = item

        return transaction_info

    def update_packages(self, packages):
        """
        Install the newest versions of the given packages.

        @param packages list of packages (names) to be updated
        """
        self._prepare_update(packages)

        pkgs = []
        for package in self.yum.tsInfo.getMembers():
            if len(package.updates) == 0:
                continue
            pkgs.append(package.name)
        if len(pkgs) > 0:
            log.info("updating packages: %s" % (" ".join(pkgs)))
        else:
            log.info("no packages to update")

        self.yum.processTransaction()

    def print_package_details(self, package):
        """
        Display detailed package information.

        @param package name of the package to display information on
        """
        subprocess.check_call(["rpm", "-qi", package])

    def get_compatible_repos_list(self):
        repos_set = set()
        pl = self.yum.doPackageLists("all")
        for p in pl.installed:
            repos_set.add(self._installed_pkg_repo(p))
        for p in pl.available:
            repos_set.add(p.repoid)
        res = []
        for r1 in repos_set:
            for r2 in repos_set:
                res.append((r1, r2, self._check_repo_compatibility(r1, r2, self._unsafe)))
        return res

    def check(self, full):
        """ Checks the local rpmdb and produces information on any problems it finds. """
        args = ["yum", "check"]
        if not full:
            args += ['dependencies', 'duplicates', 'obsoletes']
        else:
            args += ['all']
        subprocess.check_call(args)

    @contextmanager
    def locking(self):
        try:
            with discard_stdout():
                self.yum.doLock()
        except yum.Errors.LockError:
            raise pum_errors.PMLockError()

        try:
            yield
        finally:
            with discard_stdout():
                self.yum.closeRpmDB()  # like yum does
                self.yum.doUnlock()

# vim: ts=4 sts=4 sw=4 et :
